<?xml version="1.0"?>
<!DOCTYPE gui SYSTEM "kpartgui.dtd">
<gui name="dolphin" version="41">
    <MenuBar>
        <Menu name="file">
            <Action name="new_menu" />
            <Action name="file_new" />
            <Action name="new_tab" />
            <Action name="file_close" />
            <Action name="undo_close_tab" />
            <Separator/>
            <Action name="add_to_places" />
            <Separator/>
            <Action name="renamefile" />
            <Action name="duplicate" />
            <Action name="movetotrash" />
            <Action name="deletefile" />
            <Separator/>
            <Action name="show_target" />
            <Separator/>
            <Action name="properties" />
        </Menu>
        <Menu name="edit">
            <Action name="edit_undo" />
            <Separator />
            <Action name="edit_cut" />
            <Action name="edit_copy" />
            <Action name="copy_location" />
            <Action name="edit_paste" />
            <Separator />
            <Action name="show_filter_bar" />
            <Action name="edit_find" />
            <Separator />
            <Action name="toggle_selection_mode" />
            <Action name="copy_to_inactive_split_view" />
            <Action name="move_to_inactive_split_view" />
            <Action name="edit_select_all" />
            <Action name="invert_selection" />
        </Menu>
        <Menu name="view">
            <Action name="view_zoom_in"/>
            <Action name="view_zoom_reset"/>
            <Action name="view_zoom_out"/>
            <Separator/>
            <Action name="sort" />
            <Action name="view_mode" />
            <Action name="additional_info" />
            <Action name="show_preview" />
            <Action name="show_in_groups" />
            <Action name="show_hidden_files" />
            <Action name="act_as_admin" />
            <Separator/>
            <Action name="split_view_menu" />
            <Action name="popout_split_view" />
            <Action name="split_stash" />
            <Action name="redisplay" />
            <Action name="stop" />
            <Separator/>
            <Action name="panels" />
            <Menu name="location_bar" icon="edit-select-text">
                <text context="@title:menu">Location Bar</text>
                <Action name="editable_location" />
                <Action name="replace_location" />
            </Menu>
            <Separator/>
            <Action name="view_properties" />
        </Menu>
        <Menu name="go">
            <Action name="bookmarks" />
            <Action name="closed_tabs" />
        </Menu>
        <Menu name="tools">
            <Action name="open_preferred_search_tool" />
            <Action name="open_terminal" />
            <Action name="open_terminal_here" />
            <Action name="compare_files" />
            <Action name="change_remote_encoding" />
        </Menu>
    </MenuBar>
    <State name="new_file" >
        <disable>
            <Action name="edit_undo" />
            <Action name="edit_redo" />
            <Action name="edit_cut" />
            <Action name="renamefile" />
            <Action name="movetotrash" />
            <Action name="deletefile" />
            <Action name="invert_selection" />
            <Separator/>
            <Action name="go_back" />
            <Action name="go_forward" />
        </disable>
    </State>
    <State name="has_selection" >
        <enable>
            <Action name="invert_selection" />
        </enable>
    </State>
    <State name="has_no_selection" >
        <disable>
            <Action name="delete_shortcut" />
            <Action name="invert_selection" />
        </disable>
    </State>
    <ToolBar noMerge="1" name="mainToolBar" >
        <text context="@title:menu">Main Toolbar</text>
        <Action name="go_back" />
        <Action name="go_forward" />
        <Separator name="separator_1" />
        <Action name="icons" />
        <Action name="compact" />
        <Action name="details" />
        <Action name="url_navigators" />
        <Action name="split_view" />
        <Action name="split_stash" />
        <Action name="toggle_search" />
        <Action name="hamburger_menu" />
    </ToolBar>
    <ActionProperties scheme="Default">
        <Action priority="0" name="go_back"/>
        <Action priority="0" name="go_forward"/>
        <Action priority="0" name="go_up"/>
        <Action priority="0" name="go_home"/>
        <Action priority="0" name="stop"/>
        <Action priority="0" name="icons"/>
        <Action priority="0" name="compact"/>
        <Action priority="0" name="details"/>
        <Action priority="0" name="view_zoom_in"/>
        <Action priority="0" name="view_zoom_reset"/>
        <Action priority="0" name="view_zoom_out"/>
        <Action priority="0" name="edit_cut"/>
        <Action priority="0" name="edit_copy"/>
        <Action priority="0" name="edit_paste"/>
        <Action priority="0" name="toggle_search"/>
        <Action priority="0" name="toggle_filter"/>
    </ActionProperties>
</gui>
