/*
 * SPDX-FileCopyrightText: 2006 Peter Penz <peter.penz@gmx.at>
 *
 * SPDX-License-Identifier: GPL-2.0-or-later
 */

#include "settingspagebase.h"

SettingsPageBase::SettingsPageBase(QWidget *parent)
    : QWidget(parent)
{
}

SettingsPageBase::~SettingsPageBase()
{
}

#include "moc_settingspagebase.cpp"
